package com.sun.media.sound;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiDeviceTransmitter;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiDeviceReceiver;
import javax.sound.midi.Transmitter;
import java.util.Collections;
import javax.sound.midi.MidiUnavailableException;
import java.util.List;
import javax.sound.midi.Receiver;
import java.util.ArrayList;
import javax.sound.midi.MidiDevice;

abstract class AbstractMidiDevice implements MidiDevice, ReferenceCountingDevice
{
    private static final boolean TRACE_TRANSMITTER = false;
    private ArrayList<Receiver> receiverList;
    private AbstractMidiDevice.TransmitterList transmitterList;
    private final Object traRecLock;
    private final MidiDevice.Info info;
    private volatile boolean open;
    private int openRefCount;
    private List openKeepingObjects;
    protected volatile long id;
    
    protected AbstractMidiDevice(MidiDevice.Info info) {
        this.traRecLock = new Object();
        this.info = info;
        this.openRefCount = 0;
    }
    
    @Override
    public final MidiDevice.Info getDeviceInfo() {
        return this.info;
    }
    
    @Override
    public final void open() throws MidiUnavailableException {
        synchronized (this) {
            this.openRefCount = -1;
            this.doOpen();
        }
    }
    
    private void openInternal(Object o) throws MidiUnavailableException {
        synchronized (this) {
            if (this.openRefCount != -1) {
                ++this.openRefCount;
                this.getOpenKeepingObjects().add(o);
            }
            this.doOpen();
        }
    }
    
    private void doOpen() throws MidiUnavailableException {
        synchronized (this) {
            if (!this.isOpen()) {
                this.implOpen();
                this.open = true;
            }
        }
    }
    
    @Override
    public final void close() {
        synchronized (this) {
            this.doClose();
            this.openRefCount = 0;
        }
    }
    
    public final void closeInternal(Object o) {
        synchronized (this) {
            if (this.getOpenKeepingObjects().remove(o) && this.openRefCount > 0) {
                --this.openRefCount;
                if (this.openRefCount == 0) {
                    this.doClose();
                }
            }
        }
    }
    
    public final void doClose() {
        synchronized (this) {
            if (this.isOpen()) {
                this.implClose();
                this.open = false;
            }
        }
    }
    
    @Override
    public final boolean isOpen() {
        return this.open;
    }
    
    protected void implClose() {
        synchronized (this.traRecLock) {
            if (this.receiverList != null) {
                for (int i = 0; i < this.receiverList.size(); ++i) {
                    ((Receiver)this.receiverList.get(i)).close();
                }
                this.receiverList.clear();
            }
            if (this.transmitterList != null) {
                this.transmitterList.close();
            }
        }
    }
    
    @Override
    public long getMicrosecondPosition() {
        return -1L;
    }
    
    @Override
    public final int getMaxReceivers() {
        if (this.hasReceivers()) {
            return -1;
        }
        return 0;
    }
    
    @Override
    public final int getMaxTransmitters() {
        if (this.hasTransmitters()) {
            return -1;
        }
        return 0;
    }
    
    @Override
    public final Receiver getReceiver() throws MidiUnavailableException {
        Receiver receiver;
        synchronized (this.traRecLock) {
            receiver = this.createReceiver();
            this.getReceiverList().add(receiver);
        }
        return receiver;
    }
    
    @Override
    public final List<Receiver> getReceivers() {
        List<Object> list;
        synchronized (this.traRecLock) {
            if (this.receiverList == null) {
                list = Collections.unmodifiableList(new ArrayList(0));
            }
            else {
                list = Collections.unmodifiableList((List)(List)this.receiverList.clone());
            }
        }
        return list;
    }
    
    @Override
    public final Transmitter getTransmitter() throws MidiUnavailableException {
        Transmitter transmitter;
        synchronized (this.traRecLock) {
            transmitter = this.createTransmitter();
            AbstractMidiDevice.TransmitterList.access$000(this.getTransmitterList(), transmitter);
        }
        return transmitter;
    }
    
    @Override
    public final List<Transmitter> getTransmitters() {
        List<Object> list;
        synchronized (this.traRecLock) {
            if (this.transmitterList == null || this.transmitterList.transmitters.size() == 0) {
                list = Collections.unmodifiableList(new ArrayList(0));
            }
            else {
                list = Collections.unmodifiableList((List)(List)this.transmitterList.transmitters.clone());
            }
        }
        return list;
    }
    
    final long getId() {
        return this.id;
    }
    
    @Override
    public final Receiver getReceiverReferenceCounting() throws MidiUnavailableException {
        Receiver receiver;
        synchronized (this.traRecLock) {
            receiver = this.getReceiver();
            this.openInternal(receiver);
        }
        return receiver;
    }
    
    @Override
    public final Transmitter getTransmitterReferenceCounting() throws MidiUnavailableException {
        Transmitter transmitter;
        synchronized (this.traRecLock) {
            transmitter = this.getTransmitter();
            this.openInternal(transmitter);
        }
        return transmitter;
    }
    
    private synchronized List getOpenKeepingObjects() {
        if (this.openKeepingObjects == null) {
            this.openKeepingObjects = new ArrayList();
        }
        return this.openKeepingObjects;
    }
    
    private List<Receiver> getReceiverList() {
        synchronized (this.traRecLock) {
            if (this.receiverList == null) {
                this.receiverList = new ArrayList();
            }
        }
        return this.receiverList;
    }
    
    protected boolean hasReceivers() {
        return false;
    }
    
    protected Receiver createReceiver() throws MidiUnavailableException {
        throw new MidiUnavailableException("MIDI IN receiver not available");
    }
    
    final AbstractMidiDevice.TransmitterList getTransmitterList() {
        synchronized (this.traRecLock) {
            if (this.transmitterList == null) {
                this.transmitterList = new AbstractMidiDevice.TransmitterList();
            }
        }
        return this.transmitterList;
    }
    
    protected boolean hasTransmitters() {
        return false;
    }
    
    protected Transmitter createTransmitter() throws MidiUnavailableException {
        throw new MidiUnavailableException("MIDI OUT transmitter not available");
    }
    
    protected abstract void implOpen() throws MidiUnavailableException;
    
    @Override
    protected final void finalize() {
        this.close();
    }
    
    static List access$300(AbstractMidiDevice abstractMidiDevice) {
        return abstractMidiDevice.getReceiverList();
    }
    
    abstract class AbstractReceiver implements MidiDeviceReceiver
    {
        private volatile boolean open;
        
        AbstractReceiver() {
            this.open = true;
        }
        
        @Override
        public final synchronized void send(MidiMessage midiMessage, long n) {
            if (!this.open) {
                throw new IllegalStateException("Receiver is not open");
            }
            this.implSend(midiMessage, n);
            return;
        }
        
        abstract void implSend(MidiMessage p0, long p1);
        
        @Override
        public final void close() {
            this.open = false;
            synchronized (AbstractMidiDevice.this.traRecLock) {
                AbstractMidiDevice.access$300(AbstractMidiDevice.this).remove(this);
            }
            AbstractMidiDevice.this.closeInternal(this);
        }
        
        @Override
        public final MidiDevice getMidiDevice() {
            return AbstractMidiDevice.this;
        }
        
        final boolean isOpen() {
            return this.open;
        }
    }
    
    class BasicTransmitter implements MidiDeviceTransmitter
    {
        private Receiver receiver;
        AbstractMidiDevice.TransmitterList tlist;
        
        protected BasicTransmitter() {
            this.receiver = null;
            this.tlist = null;
        }
        
        private void setTransmitterList(AbstractMidiDevice.TransmitterList tlist) {
            this.tlist = tlist;
        }
        
        @Override
        public final void setReceiver(Receiver receiver) {
            if (this.tlist != null && this.receiver != receiver) {
                AbstractMidiDevice.TransmitterList.access$400(this.tlist, this, this.receiver, receiver);
                this.receiver = receiver;
            }
        }
        
        @Override
        public final Receiver getReceiver() {
            return this.receiver;
        }
        
        @Override
        public final void close() {
            AbstractMidiDevice.this.closeInternal(this);
            if (this.tlist != null) {
                AbstractMidiDevice.TransmitterList.access$400(this.tlist, this, this.receiver, null);
                AbstractMidiDevice.TransmitterList.access$500(this.tlist, this);
                this.tlist = null;
            }
        }
        
        @Override
        public final MidiDevice getMidiDevice() {
            return AbstractMidiDevice.this;
        }
        
        static void access$600(AbstractMidiDevice.BasicTransmitter basicTransmitter, AbstractMidiDevice.TransmitterList transmitterList) {
            basicTransmitter.setTransmitterList(transmitterList);
        }
    }
    
    final class TransmitterList
    {
        private final ArrayList<Transmitter> transmitters;
        private MidiOutDevice.MidiOutReceiver midiOutReceiver;
        private int optimizedReceiverCount;
        
        TransmitterList() {
            this.transmitters = new ArrayList();
            this.optimizedReceiverCount = 0;
        }
        
        private void add(Transmitter e) {
            synchronized (this.transmitters) {
                this.transmitters.add(e);
            }
            if ((e instanceof AbstractMidiDevice.BasicTransmitter)) {
                AbstractMidiDevice.BasicTransmitter.access$600((AbstractMidiDevice.BasicTransmitter)e, this);
            }
        }
        
        private void remove(Transmitter o) {
            synchronized (this.transmitters) {
                int index = this.transmitters.indexOf(o);
                if (index >= 0) {
                    this.transmitters.remove(index);
                }
            }
        }
        
        private void receiverChanged(AbstractMidiDevice.BasicTransmitter basicTransmitter, Receiver receiver, Receiver receiver2) {
            synchronized (this.transmitters) {
                if (this.midiOutReceiver == receiver) {
                    this.midiOutReceiver = null;
                }
                if (receiver2 != null && receiver2 instanceof MidiOutDevice.MidiOutReceiver && this.midiOutReceiver == null) {
                    this.midiOutReceiver = (MidiOutDevice.MidiOutReceiver)receiver2;
                }
                this.optimizedReceiverCount = (this.midiOutReceiver != null);
            }
        }
        
        void close() {
            synchronized (this.transmitters) {
                for (int i = 0; i < this.transmitters.size(); ++i) {
                    ((Transmitter)this.transmitters.get(i)).close();
                }
                this.transmitters.clear();
            }
        }
        
        void sendMessage(int n, long n2) {
            try {
                synchronized (this.transmitters) {
                    int size = this.transmitters.size();
                    if (this.optimizedReceiverCount == size) {
                        if (this.midiOutReceiver != null) {
                            this.midiOutReceiver.sendPackedMidiMessage(n, n2);
                        }
                    }
                    else {
                        for (int i = 0; i < size; ++i) {
                            Receiver receiver = ((Transmitter)this.transmitters.get(i)).getReceiver();
                            if (receiver != null) {
                                if (this.optimizedReceiverCount > 0) {
                                    if ((receiver instanceof MidiOutDevice.MidiOutReceiver)) {
                                        ((MidiOutDevice.MidiOutReceiver)receiver).sendPackedMidiMessage(n, n2);
                                    }
                                    else {
                                        receiver.send(new FastShortMessage(n), n2);
                                    }
                                }
                                else {
                                    receiver.send(new FastShortMessage(n), n2);
                                }
                            }
                        }
                    }
                }
            }
            catch (InvalidMidiDataException ex) {}
        }
        
        void sendMessage(byte[] array, long n) {
            try {
                synchronized (this.transmitters) {
                    for (int size = this.transmitters.size(), i = 0; i < size; ++i) {
                        Receiver receiver = ((Transmitter)this.transmitters.get(i)).getReceiver();
                        if (receiver != null) {
                            receiver.send(new FastSysexMessage(array), n);
                        }
                    }
                }
            }
            catch (InvalidMidiDataException ex) {
                return;
            }
        }
        
        void sendMessage(MidiMessage midiMessage, long n) {
            if ((midiMessage instanceof FastShortMessage)) {
                this.sendMessage(((FastShortMessage)midiMessage).getPackedMsg(), n);
                return;
            }
            synchronized (this.transmitters) {
                int size = this.transmitters.size();
                if (this.optimizedReceiverCount == size) {
                    if (this.midiOutReceiver != null) {
                        this.midiOutReceiver.send(midiMessage, n);
                    }
                }
                else {
                    for (int i = 0; i < size; ++i) {
                        Receiver receiver = ((Transmitter)this.transmitters.get(i)).getReceiver();
                        if (receiver != null) {
                            receiver.send(midiMessage, n);
                        }
                    }
                }
            }
            return;
        }
        
        static void access$000(AbstractMidiDevice.TransmitterList list, Transmitter transmitter) {
            list.add(transmitter);
        }
        
        static void access$400(AbstractMidiDevice.TransmitterList list, AbstractMidiDevice.BasicTransmitter basicTransmitter, Receiver receiver, Receiver receiver2) {
            list.receiverChanged(basicTransmitter, receiver, receiver2);
        }
        
        static void access$500(AbstractMidiDevice.TransmitterList list, Transmitter transmitter) {
            list.remove(transmitter);
        }
    }
}
